/** @internal
 ** @file     curvdsift.c
 ** @author   Philipp Fischer
 ** @brief    Modified DSIFT MEX to include curvature
 **/

/*
Copyright (C) 2007-12 Andrea Vedaldi and Brian Fulkerson.
All rights reserved.

This file is part of the VLFeat library and is made available under
the terms of the BSD license (see the COPYING file).
*/

#include <mexutils.h>
#include <vl/mathop.h>
#include <vl/curvdsift.h>

#include <math.h>
#include <assert.h>

/* option codes */
enum {
  opt_step = 0,
  opt_bounds,
  opt_size,
  opt_fast,
  opt_norm,
  opt_window_size,
  opt_float_descriptors,
  opt_verbose,
  opt_curvsettings
} ;

/* options */
vlmxOption  options [] = {
{"Bounds",           1,   opt_bounds           },
{"Step",             1,   opt_step             },
{"Size",             1,   opt_size             },
{"Fast",             0,   opt_fast             },
{"Norm",             0,   opt_norm             },
{"WindowSize",       1,   opt_window_size      },
{"FloatDescriptors", 0,   opt_float_descriptors},
{"Verbose",          0,   opt_verbose          },
{"Curvsettings",     1,   opt_curvsettings     },
{0,                  0,   0                    }
} ;

/** ------------------------------------------------------------------
 ** @brief MEX entry point
 **/

void
mexFunction(int nout, mxArray *out[], 
            int nin, const mxArray *in[])
{
  enum {IN_I=0, IN_END} ;
  enum {OUT_FRAMES=0, OUT_DESCRIPTORS} ;

  int                verbose = 0 ;
  int                opt ;
  int                next = IN_END ;
  mxArray const     *optarg ;

  float const       *data ;
  int                M, N ;

  int                step = 1 ;
  int                size = 3 ;
  vl_bool            norm = 0 ;

  vl_bool floatDescriptors = VL_FALSE ;
  vl_bool useFlatWindow = VL_FALSE ;
  double windowSize = -1.0 ;

  double *bounds = NULL ;
  double boundBuffer [4] ;

  VL_USE_MATLAB_ENV ;

  /* -----------------------------------------------------------------
   *                                               Check the arguments
   * -------------------------------------------------------------- */

  if (nin < 1) {
    vlmxError(vlmxErrNotEnoughInputArguments, NULL) ;
  } else if (nout > 3) {
    vlmxError(vlmxErrTooManyOutputArguments, NULL) ;
  }

  if (mxGetNumberOfDimensions (in[IN_I]) != 2              ||
      mxGetClassID            (in[IN_I]) != mxSINGLE_CLASS ) {
    vlmxError(vlmxErrInvalidArgument,
              "I must be a matrix of class SINGLE.") ;
  }

  data = (float*) mxGetData (in[IN_I]) ;
  M    = mxGetM (in[IN_I]) ;
  N    = mxGetN (in[IN_I]) ;

  char curvsettings[129] = {0};
  
  while ((opt = vlmxNextOption (in, nin, options, &next, &optarg)) >= 0) {
    switch (opt) {

      case opt_verbose :
        ++ verbose ;
        break ;

      case opt_fast :
        useFlatWindow = 1 ;
        break ;

      case opt_norm :
        norm = 1 ;
        break ;

      case opt_bounds :
        if (!vlmxIsPlainVector(optarg, 4)) {
          mexErrMsgTxt("BOUNDS must be a 4-dimensional vector.") ;
        }
        bounds = boundBuffer ;
        bounds [0] = mxGetPr(optarg)[0] - 1 ;
        bounds [1] = mxGetPr(optarg)[1] - 1 ;
        bounds [2] = mxGetPr(optarg)[2] - 1 ;
        bounds [3] = mxGetPr(optarg)[3] - 1 ;
        break ;

      case opt_size :
        if (!vlmxIsPlainScalar(optarg) || (size = (int) *mxGetPr(optarg)) <= 0) {
          vlmxError(vlmxErrInvalidArgument,"SIZE is not a scalar or it is negative.") ;
        }
        break ;

      case opt_step :
        if (!vlmxIsPlainScalar(optarg) || (step = (int) *mxGetPr(optarg)) <= 0) {
          vlmxError(vlmxErrInvalidArgument,"STEP is not a scalar or it is negative.") ;
        }
        break ;

      case opt_window_size :
        if (!vlmxIsPlainScalar(optarg) || (windowSize = *mxGetPr(optarg)) < 0) {
          vlmxError(vlmxErrInvalidArgument,"WINDOWSIZE is not a scalar or it is negative.") ;
        }
        break ;

      case opt_float_descriptors :
        floatDescriptors = VL_TRUE ;
        break ;

      case opt_curvsettings :
        if (!vlmxIsString(optarg, -1)) {
          vlmxError(vlmxErrInvalidArgument,"Curvsettings is not a string.") ;
        }
        mxGetString(optarg, curvsettings, 128) ;
        curvsettings[128 - 1] = 0 ;
        
      break ;
        
      default :
        abort() ;
    }
  }

  /* -----------------------------------------------------------------
   *                                                            Do job
   * -------------------------------------------------------------- */
  {
    int numFrames ;
    int descrSizeA, descrSizeB, descrSizeBoth ;
    VlDsiftKeypoint const *frames ;
    VlDsiftDescriptorGeometry const *geom ;
    float const *descrsA, *descrsB ;
    int k, i ;

    VlDsiftFilter *dsift ;

    /* note that the image received from MATLAB is transposed */
    dsift = vl_curvdsift_new_basic (M, N, step, size, curvsettings) ;
    
    if (bounds) {
      vl_curvdsift_set_bounds(dsift,
                          VL_MAX(bounds[1], 0),
                          VL_MAX(bounds[0], 0),
                          VL_MIN(bounds[3], M - 1),
                          VL_MIN(bounds[2], N - 1));
    }
    vl_curvdsift_set_flat_window(dsift, useFlatWindow) ;

    if (windowSize >= 0) {
      vl_curvdsift_set_window_size(dsift, windowSize) ;
    }

    numFrames = vl_curvdsift_get_keypoint_num (dsift) ;
    descrSizeA = vl_curvdsift_get_descriptor_size (dsift, 0) ;
    descrSizeB = vl_curvdsift_get_descriptor_size (dsift, 1) ;
    descrSizeBoth = descrSizeA + descrSizeB;
    
    geom = vl_curvdsift_get_geometry (dsift) ;

    if (verbose) {
      int stepX ;
      int stepY ;
      int minX ;
      int minY ;
      int maxX ;
      int maxY ;
      vl_bool useFlatWindow ;

      vl_curvdsift_get_steps (dsift, &stepY, &stepX) ;
      vl_curvdsift_get_bounds (dsift, &minY, &minX, &maxY, &maxX) ;
      useFlatWindow = vl_curvdsift_get_flat_window(dsift) ;

      mexPrintf("vl_dsift: image size         [W, H] = [%d, %d]\n", N, M) ;
      mexPrintf("vl_dsift: bounds:            [minX,minY,maxX,maxY] = [%d, %d, %d, %d]\n",
                minX+1, minY+1, maxX+1, maxY+1) ;
      mexPrintf("vl_dsift: subsampling steps: stepX=%d, stepY=%d\n", stepX, stepY) ;
      mexPrintf("vl_dsift: num bins:          [numBinT, numBinX, numBinY, numCurv] = [%d, %d, %d, %d]\n",
                geom->numBinT,
                geom->numBinX,
                geom->numBinY,
                dsift->numBinCurv
               ) ;
      mexPrintf("vl_dsift: descriptor size:   %d,%d\n", descrSizeA, descrSizeB) ;
      mexPrintf("vl_dsift: bin sizes:         [binSizeX, binSizeY] = [%d, %d]\n",
                geom->binSizeX,
                geom->binSizeY) ;
      mexPrintf("vl_dsift: flat window:       %s\n", VL_YESNO(useFlatWindow)) ;
      mexPrintf("vl_dsift: window size:       %g\n", vl_curvdsift_get_window_size(dsift)) ;
      mexPrintf("vl_dsift: num of features:   %d\n", numFrames) ;
      mexPrintf("vl_dsift: curvsettings:      %s\n", curvsettings) ;
    }

    vl_curvdsift_process (dsift, data) ;

    frames = vl_curvdsift_get_keypoints (dsift) ; 
    descrsA = vl_curvdsift_get_descriptors (dsift, 0);
    descrsB = vl_curvdsift_get_descriptors (dsift, 1);

    /* ---------------------------------------------------------------
     *                                            Create output arrays
     * ------------------------------------------------------------ */
    {
      mwSize dims [2] ;

      dims [0] = descrSizeBoth ;
      dims [1] = numFrames ;

      if (floatDescriptors) {
        out[OUT_DESCRIPTORS] = mxCreateNumericArray
        (2, dims, mxSINGLE_CLASS, mxREAL) ;
      } else {
        out[OUT_DESCRIPTORS] = mxCreateNumericArray
        (2, dims, mxUINT8_CLASS, mxREAL) ;
      }

      dims [0] = norm ? 3 : 2 ;

      out[OUT_FRAMES] = mxCreateNumericArray
      (2, dims, mxDOUBLE_CLASS, mxREAL) ;
    } 

    /* ---------------------------------------------------------------
     *                                          Curvature Debug Output
     * ------------------------------------------------------------ */
    if(nout >= 3)
    {
      mwSize pdims [3] ;
        pdims [0] = M;
        pdims [1] = N;
        pdims [2] = 12 + 5;
      
      out[2] = mxCreateNumericArray(3, pdims, mxSINGLE_CLASS, mxREAL) ;
      float *outDbg = mxGetData(out[2]) ;
      
      for(int t = 0; t < 12+5; t++) {
        vl_curvdsift_get_dbg(dsift, outDbg + (M*N*t), t);
      }

    }
    
    /* ---------------------------------------------------------------
     *                                                       Copy back
     * ------------------------------------------------------------ */
    { 
      float *tmpDescr = mxMalloc(sizeof(float) * descrSizeBoth) ;
      double *outFrameIter = mxGetPr(out[OUT_FRAMES]) ;
      void *outDescrIter = mxGetData(out[OUT_DESCRIPTORS]) ;
      
      for (k = 0 ; k < numFrames ; ++k) {
        *outFrameIter++ = frames[k].y + 1 ;
        *outFrameIter++ = frames[k].x + 1 ;

        /* We have an implied / 2 in the norm, because of the clipping
           below */
        if (norm)
          *outFrameIter++ = frames [k].norm ;

        vl_curvdsift_transpose_descriptor (tmpDescr,
                                       descrsA + descrSizeA * k,
                                       geom->numBinT,
                                       geom->numBinX,
                                       geom->numBinY) ;
        
        vl_curvdsift_transpose_curvdescriptor (dsift, tmpDescr + descrSizeA,
                                       descrsB + descrSizeB * k,
                                       geom->numBinX,
                                       geom->numBinY) ;
        
        if (floatDescriptors) {
          for (i = 0 ; i < descrSizeBoth ; ++i) {
            float * pt = (float*) outDescrIter ;
            *pt++ = VL_MIN(512.0F * tmpDescr[i], 255.0F) ;
            outDescrIter = pt ;
          }
        } else {
          for (i = 0 ; i < descrSizeBoth ; ++i) {
            vl_uint8 * pt = (vl_uint8*) outDescrIter ;
            *pt++ = (vl_uint8) (VL_MIN(512.0F * tmpDescr[i], 255.0F)) ;
            outDescrIter = pt ;

          }
        }
      }
      mxFree(tmpDescr) ;
    }
    vl_curvdsift_delete (dsift) ;
  }
}
