/** @file curvdsift.c
 ** @brief Dense SIFT - Definition
 ** @author Andrea Vedaldi
 ** @author Brian Fulkerson
 ** @author Curvature Integration by Philipp Fischer
 **/

/*
Copyright (C) 2007-12 Andrea Vedaldi and Brian Fulkerson.
All rights reserved.

This file is part of the VLFeat library and is made available under
the terms of the BSD license (see the COPYING file).
*/

#include "curvdsift.h"
#include "pgm.h"
#include "mathop.h"
#include "imopv.h"
#include <math.h>

#define eps 0.0001

#define NELEMS(x)  (sizeof(x) / sizeof(x[0]))

/**
<!-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  -->
@page dsift Dense Scale Invariant Feature Transform (DSIFT)
@author Andrea Vedaldi
@author Brian Fulkerson
<!-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  -->

@ref dsift.h implements a dense version of @ref sift.h "SIFT". This is
an object that can quickly compute descriptors for densely sampled
keypoints with identical size and orientation. It can be reused for
multiple images of the same size.

- @ref dsift-intro
- @ref dsift-usage
- @ref dsift-tech
  - @ref dsift-tech-descriptor-dense
  - @ref dsift-tech-sampling

<!-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  -->
@section dsift-intro Overview
<!-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  -->

@sa @ref sift "The SIFT module", @ref dsift-tech "Technical details"

This module implements a fast algorithm for the calculation of a large
number of SIFT descriptors of densely sampled features of the same
scale and orientation. See the @ref sift "SIFT module" for an
overview of SIFT.

The feature frames (keypoints) are indirectly specified by the
sampling steps (::vl_curvdsift_set_steps) and the sampling bounds
(::vl_curvdsift_set_bounds).  The descriptor geometry (number and size of
the spatial bins and number of orientation bins) can be customized
(::vl_curvdsift_set_geometry, ::VlDsiftDescriptorGeometry).

@image html dsift-geom.png "Dense SIFT descriptor geometry"

By default, SIFT uses a Gaussian windowing function that discounts
contributions of gradients further away from the descriptor
centers. This function can be changed to a flat window by invoking
::vl_curvdsift_set_flat_window. In this case, gradients are accumulated
using only bilinear interpolation, but instad of being reweighted by a
Gassuain window, they are all weighted equally. However, after
gradients have been accumulated into a spatial bin, the whole bin is
reweighted by the average of the Gaussian window over the spatial
support of that bin. This &ldquo;approximation&rdquo; substantially
improves speed with little or no loss of performance in applications.

Keypoints are sampled in such a way that the centers of the spatial
bins are at integer coordinates within the image boundaries. For
instance, the top-left bin of the top-left descriptor is centered on
the pixel (0,0). The bin immediately to the right at
(<code>binSizeX</code>,0), where <code>binSizeX</code> is a paramtere
in the ::VlDsiftDescriptorGeometry structure. ::vl_curvdsift_set_bounds
can be used to further restrict sampling to the keypoints in an image.

<!-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  -->
 @section dsift-usage Usage
<!-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  -->

DSIFT is implemented by a ::VlDsiftFilter object that can be used
to process a sequence of images of a given geometry.
To use the <b>DSIFT filter</b>:

- Initialize a new DSIFT filter object by ::vl_curvdsift_new (or the simplified
::vl_curvdsift_new_basic). Customize the descriptor parameters by
::vl_curvdsift_set_steps, ::vl_curvdsift_set_geometry, etc.
- Process an image by ::vl_curvdsift_process.
- Retrieve the number of keypoints (::vl_curvdsift_get_keypoint_num), the
  keypoints (::vl_curvdsift_get_keypoints), and their descriptors
  (::vl_curvdsift_get_descriptors).
- Optionally repeat for more images.
- Delete the DSIFT filter by ::vl_curvdsift_delete.

<!-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  -->
@section dsift-tech Technical details
<!-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  -->

This section extends the @ref sift-tech-descriptor "SIFT descriptor section"
and specialzies it to the case of dense keypoints.

<!-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  -->
@subsection dsift-tech-descriptor-dense Dense descriptors
<!-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  -->

When computing descriptors for many keypoints differing only by their
position (and with null rotation), further simplifications are
possible. In this case, in fact,

@f{eqnarray*}
     \mathbf{x} &=& m \sigma \hat{\mathbf{x}} + T,\\
 h(t,i,j)
 &=&
 m \sigma \int
 g_{\sigma_\mathrm{win}}(\mathbf{x} - T)\,
 w_\mathrm{ang}(\angle J(\mathbf{x}) - \theta_t)\,
 w\left(\frac{x - T_x}{m\sigma} - \hat{x}_i\right)\,
 w\left(\frac{y - T_y}{m\sigma} - \hat{y}_j\right)\,
 |J(\mathbf{x})|\,
 d\mathbf{x}.
@f}

Since many different values of @e T are sampled, this is conveniently
expressed as a separable convolution. First, we translate by @f$
\mathbf{x}_{ij} = m\sigma(\hat x_i,\ \hat y_i)^\top @f$ and we use the
symmetry of the various binning and windowing functions to write

@f{eqnarray*}
 h(t,i,j)
 &=&
 m \sigma \int
 g_{\sigma_\mathrm{win}}(T' - \mathbf{x} - \mathbf{x}_{ij})\,
 w_\mathrm{ang}(\angle J(\mathbf{x}) - \theta_t)\,
 w\left(\frac{T'_x - x}{m\sigma}\right)\,
 w\left(\frac{T'_y - y}{m\sigma}\right)\,
 |J(\mathbf{x})|\,
 d\mathbf{x},
\\
T' &=& T + m\sigma
\left[\begin{array}{cc} x_i \\ y_j \end{array}\right].
@f}

Then we define kernels

@f{eqnarray*}
 k_i(x) &=&
 \frac{1}{\sqrt{2\pi} \sigma_{\mathrm{win}}}
 \exp\left(
 -\frac{1}{2}
 \frac{(x-x_i)^2}{\sigma_{\mathrm{win}}^2}
 \right)
 w\left(\frac{x}{m\sigma}\right),
 \\
 k_j(y) &=&
 \frac{1}{\sqrt{2\pi} \sigma_{\mathrm{win}}}
 \exp\left(
 -\frac{1}{2}
 \frac{(y-y_j)^2}{\sigma_{\mathrm{win}}^2}
 \right)
 w\left(\frac{y}{m\sigma}\right),
@f}

and obtain

@f{eqnarray*}
 h(t,i,j) &=& (k_ik_j * \bar J_t)\left( T + m\sigma
\left[\begin{array}{cc} x_i \\ y_j \end{array}\right] \right),
\\
\bar J_t(\mathbf{x}) &=&  w_\mathrm{ang}(\angle J(\mathbf{x}) - \theta_t)\,|J(\mathbf{x})|.
@f}

Furthermore, if we use a flat rather than Gaussian windowing function,
the kernels do not depend on the bin, and we have

@f{eqnarray*}
 k(z) &=&
 \frac{1}{\sigma_{\mathrm{win}}}
 w\left(\frac{z}{m\sigma}\right),
\\
 h(t,i,j) &=& (k(x)k(y) * \bar J_t)\left( T + m\sigma
\left[\begin{array}{cc} x_i \\ y_j \end{array}\right] \right),
@f}

(here @f$ \sigma_\mathrm{win} @f$ is the side of the flat window).

@note In this case the binning functions @f$ k(z) @f$ are triangular
and the convolution can be computed in time independent on the filter
(i.e. descriptor bin) support size by integral signals.

<!-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  -->
@subsection dsift-tech-sampling Sampling
<!-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  -->

To avoid resampling and dealing with special boundary conditions, we
impose some mild restrictions on the geometry of the descriptors that
can be computed. In particular, we impose that the bin centers @f$ T +
m\sigma (x_i,\ y_j) @f$ are always at integer coordinates within the
image boundaries. This eliminates the need for costly interpolation.
This condition amounts to (expressed in terms of the @e x coordinate,
and equally applicable to @e y)

@f[
 \{0,\dots, W-1\} \ni T_x + m\sigma x_i =
 T_x + m\sigma i - \frac{N_x-1}{2}
 = \bar T_x + m\sigma i,
 \qquad i = 0,\dots,N_x-1.
@f]

Notice that for this condition to be satisfied, the @em descriptor
center @f$ T_x @f$ needs to be either fractional or integer depending
on @f$ N_x @f$ being even or odd. To eliminate this complication,
it is simpler to use as a reference not the descriptor center @e T,
but the coordinates of the upper-left bin @f$ \bar T @f$. Thus we
sample the latter on a regular (integer) grid

@f[
 \left[\begin{array}{cc}
   0 \\
   0
 \end{array}\right]
 \leq
 \bar T =
 \left[\begin{array}{cc}
   \bar T_x^{\min} + p \Delta_x \\
   \bar T_y^{\min} + q \Delta_y \\
 \end{array}\right]
 \leq
 \left[\begin{array}{cc}
   W - 1 - m\sigma N_x \\
   H - 1 - m\sigma N_y
 \end{array}\right],
 \quad
 \bar T =
 \left[\begin{array}{cc}
   T_x - \frac{N_x - 1}{2} \\
   T_y - \frac{N_y - 1}{2} \\
  \end{array}\right]
@f]

and we impose that the bin size @f$ m \sigma @f$ is integer as well.

**/

const double uu8[8] = {1.0000,
    0.7071,
    0.0000,
   -0.7071,
   -1.0000,
   -0.7071,
   -0.0000,
    0.7071};

const double vv8[8] = {0.0000, 
    0.7071,
    1.0000,
    0.7071,
    0.0000,
   -0.7071,
   -1.0000,
   -0.7071};


/** ------------------------------------------------------------------
 ** @internal @brief Initialize new convolution kernel
 ** @param binSize
 ** @param numBins
 ** @param binIndex negative to use flat window.
 ** @param windowSize
 ** @return a pointer to new filter.
 **/

float *
_vl_curvdsift_new_kernel (int binSize, int numBins, int binIndex, double windowSize)
{
  int filtLen = 2 * binSize - 1 ;
  float * ker = vl_malloc (sizeof(float) * filtLen) ;
  float * kerIter = ker ;
  float delta = binSize * (binIndex - 0.5F * (numBins - 1)) ;
  /*
  float sigma = 0.5F * ((numBins - 1) * binSize + 1) ;
  float sigma = 0.5F * ((numBins) * binSize) ;
  */
  float sigma = (float) binSize * (float) windowSize ;
  int x ;

  for (x = - binSize + 1 ; x <= + binSize - 1 ; ++ x) {
    float z = (x - delta) / sigma ;
    *kerIter++ = (1.0F - fabsf(x) / binSize) *
      ((binIndex >= 0) ? expf(- 0.5F * z*z) : 1.0F) ;
  }
  return ker ;
}

void vl_pimconv_vf (float *dst, float *tmp,float const *src, vl_size src_width,
vl_size src_height, float const *filt1, int s1, float const *filt2, int s2, int step)
{
  int r1 = (s1-1)/2;
  int r2 = (s2-1)/2;
  
  vl_imconvcol_vf (tmp, src_height,
                        src, src_width, src_height,
                        src_width,
                        filt1, -r1, +r1, step,
                        VL_PAD_BY_CONTINUITY|VL_TRANSPOSE) ;

  vl_imconvcol_vf (dst, src_width,
                    tmp, src_height, src_width,
                    src_height,
                    filt2, -r2, +r2, step,
                    VL_PAD_BY_CONTINUITY|VL_TRANSPOSE) ;
}

static float
_vl_curvdsift_get_bin_window_mean
(int binSize, int numBins, int binIndex, double windowSize)
{
  float delta = binSize * (binIndex - 0.5F * (numBins - 1)) ;
  /*float sigma = 0.5F * ((numBins - 1) * binSize + 1) ;*/
  float sigma = (float) binSize * (float) windowSize ;
  int x ;

  float acc = 0.0 ;
  for (x = - binSize + 1 ; x <= + binSize - 1 ; ++ x) {
    float z = (x - delta) / sigma ;
    acc += ((binIndex >= 0) ? expf(- 0.5F * z*z) : 1.0F) ;
  }
  return acc /= (2 * binSize - 1) ;
}

/** ------------------------------------------------------------------
 ** @internal @brief Normalize histogram
 ** @param begin first element of the histogram.
 ** @param end last plus one element of the histogram.
 **
 ** The function divides the specified histogram by its l2 norm.
 **/

VL_INLINE float
_vl_curvdsift_normalize_histogram (float * begin, float * end)
{
  float * iter ;
  float  norm = 0.0F ;

  for (iter = begin ; iter < end ; ++ iter) {
    norm += (*iter) * (*iter) ;
  }
  norm = vl_fast_sqrt_f (norm) + VL_EPSILON_F ;

  for (iter = begin; iter < end ; ++ iter) {
    *iter /= norm ;
  }
  return norm ;
}

/** ------------------------------------------------------------------
 ** @internal @brief Normalize histogram which is split over multiple memory areas
 **
 ** The function divides the specified histogram by its l2 norm.
 **/

VL_INLINE float
_vl_curvdsift_normalize_splithistogram (float * begin, float * end, int stride, int innercnt)
{
  float *iter, *ff;
  float  norm = 0.0F ;

  for (iter = begin ; iter < end ; iter+=stride) {
    for (ff = iter; ff < iter+innercnt; ++ff) {
      norm += (*ff) * (*ff) ;
    }
  }
  norm = vl_fast_sqrt_f (norm) + VL_EPSILON_F ;

  for (iter = begin ; iter < end ; iter+=stride) {
    for (ff = iter; ff < iter+innercnt; ++ff) {
    *ff /= norm ;
    }
  }
  
  return norm ;
}

/** ------------------------------------------------------------------
 ** @internal @brief Free internal buffers
 ** @param self DSIFT filter.
 **/

static void
_vl_curvdsift_free_buffers (VlDsiftFilter* self)
{
  if (self->frames) {
    vl_free(self->frames) ;
    self->frames = NULL ;
  }
  if (self->descrs) {
    vl_free(self->descrs) ;
    self->descrs = NULL ;
  }
  if (self->grads) {
    int t ;
    for (t = 0 ; t < self->numGradAlloc ; ++t)
      if (self->grads[t]) vl_free(self->grads[t]) ;
    vl_free(self->grads) ;
    self->grads = NULL ;
  }
  
  if (self->xxcurv) {vl_free(self->xxcurv); self->xxcurv = NULL;}
  if (self->xycurv) {vl_free(self->xycurv); self->xycurv = NULL;}
  if (self->yxcurv) {vl_free(self->yxcurv); self->yxcurv = NULL;}
  if (self->yycurv) {vl_free(self->yycurv); self->yycurv = NULL;}

  if (self->xpixcurv) {vl_free(self->xpixcurv); self->xpixcurv = NULL;}
  if (self->ypixcurv) {vl_free(self->ypixcurv); self->ypixcurv = NULL;}
  if (self->pixcurvmag) {vl_free(self->pixcurvmag); self->pixcurvmag = NULL;}
  
  if (self->gradmag) {vl_free(self->gradmag); self->gradmag = NULL;}
  if (self->binnedCurv) {
    int t ;
    for (t = 0 ; t < self->numCurvAlloc ; ++t)
      if (self->binnedCurv[t]) vl_free(self->binnedCurv[t]) ;
    vl_free(self->binnedCurv) ;
    self->binnedCurv = NULL ;
  }
  if (self->curvdescrs) {
    vl_free(self->curvdescrs) ;
    self->curvdescrs = NULL ;
  }
  
  
	  
  self->numFrameAlloc = 0 ;
  self->numBinAlloc = 0 ;
  self->numGradAlloc = 0 ;
  self->numCurvAlloc = 0 ;
  
}

/** ------------------------------------------------------------------
 ** @internal @brief Updates internal buffers to current geometry
 **/

VL_EXPORT void
_vl_curvdsift_update_buffers (VlDsiftFilter * self)
{
    if(self->curvsettings != NULL) {
        if(strHas(self->curvsettings, 'H'))  {
            self->numBinCurv = 3*8+1 ; //25
        } else {
            self->numBinCurv = 4 + 8; //12
        }
    }

    int x1 = self->boundMinX ;
  int x2 = self->boundMaxX ;
  int y1 = self->boundMinY ;
  int y2 = self->boundMaxY ;

  int rangeX = x2 - x1 - (self->geom.numBinX - 1) * self->geom.binSizeX ;
  int rangeY = y2 - y1 - (self->geom.numBinY - 1) * self->geom.binSizeY ;

  int numFramesX = (rangeX >= 0) ? rangeX / self->stepX + 1 : 0 ;
  int numFramesY = (rangeY >= 0) ? rangeY / self->stepY + 1 : 0 ;

  self->numFrames = numFramesX * numFramesY ;
  self->descrSize = self->geom.numBinT *
                    self->geom.numBinX *
                    self->geom.numBinY ;
}

/** ------------------------------------------------------------------
 ** @internal @brief Allocate internal buffers
 ** @param self DSIFT filter.
 **
 ** The function (re)allocates the internal buffers in accordance with
 ** the current image and descriptor geometry.
 **/

static void
_vl_curvdsift_alloc_buffers (VlDsiftFilter* self)
{
  _vl_curvdsift_update_buffers (self) ;
  { 
    int numFrameAlloc = vl_curvdsift_get_keypoint_num (self) ;
    int numBinAlloc   = vl_curvdsift_get_descriptor_size (self,0) ;
    int numGradAlloc  = self->geom.numBinT ;
    int numCurvAlloc  = vl_curvdsift_get_descriptor_size (self,1) ;
    
    /* see if we need to update the buffers */
    if (numBinAlloc != self->numBinAlloc ||
        numGradAlloc != self->numGradAlloc ||
        numFrameAlloc != self->numFrameAlloc ||
        numCurvAlloc != self->numCurvAlloc) {

      int t ;

      _vl_curvdsift_free_buffers(self) ;

      self->frames = vl_malloc(sizeof(VlDsiftKeypoint) * numFrameAlloc) ;
      self->descrs = vl_malloc(sizeof(float) * numBinAlloc * numFrameAlloc) ;
      self->curvdescrs = vl_malloc(sizeof(float) * numCurvAlloc * numFrameAlloc) ;
      self->grads  = vl_malloc(sizeof(float*) * numGradAlloc) ;
      for (t = 0 ; t < numGradAlloc ; ++t) {
        self->grads[t] =
          vl_malloc(sizeof(float) * self->imWidth * self->imHeight) ;
      }
      
      self->xxcurv = vl_malloc(sizeof(float) * self->imWidth * self->imHeight);
      self->xycurv = vl_malloc(sizeof(float) * self->imWidth * self->imHeight);
      self->yxcurv = vl_malloc(sizeof(float) * self->imWidth * self->imHeight);
      self->yycurv = vl_malloc(sizeof(float) * self->imWidth * self->imHeight);

      self->xpixcurv = vl_malloc(sizeof(float) * self->imWidth * self->imHeight);
      
      self->ypixcurv = vl_malloc(sizeof(float) * self->imWidth * self->imHeight);
      self->pixcurvmag = vl_malloc(sizeof(float) * self->imWidth * self->imHeight);
      
      self->gradmag = vl_malloc(sizeof(float) * self->imWidth * self->imHeight);

      self->binnedCurv = vl_malloc(sizeof(float*) * numCurvAlloc) ;
      for (t = 0 ; t < numCurvAlloc ; ++t) {
        self->binnedCurv[t] =
          vl_malloc(sizeof(float) * self->imWidth * self->imHeight) ;
          
          memset(self->binnedCurv[t],0,sizeof(float) * self->imWidth * self->imHeight); // <-- ZERO!
      }
      
      self->numBinAlloc = numBinAlloc ;
      self->numGradAlloc = numGradAlloc ;
      self->numFrameAlloc = numFrameAlloc ;
      self->numCurvAlloc = numCurvAlloc;
    }
  }
}

/** ------------------------------------------------------------------
 ** @brief Create a new DSIFT filter
 **
 ** @param imWidth width of the image.
 ** @param imHeight height of the image
 **
 ** @return new filter.
 **/

VL_EXPORT VlDsiftFilter *
vl_curvdsift_new (int imWidth, int imHeight)
{
  VlDsiftFilter * self = vl_malloc (sizeof(VlDsiftFilter)) ;
  self->imWidth  = imWidth ;
  self->imHeight = imHeight ;

  self->stepX = 5 ;
  self->stepY = 5 ;

  self->boundMinX = 0 ;
  self->boundMinY = 0 ;
  self->boundMaxX = imWidth - 1 ;
  self->boundMaxY = imHeight - 1 ;

  self->geom.numBinX = 4 ;
  self->geom.numBinY = 4 ;
  self->geom.numBinT = 8 ;
  self->geom.binSizeX = 5 ;
  self->geom.binSizeY = 5 ;
  self->numBinCurv = 0; // to be updated in update_buffers

  self->useFlatWindow = VL_FALSE ;
  self->windowSize = 2.0 ;

  self->convTmp1 = vl_malloc(sizeof(float) * self->imWidth * self->imHeight) ;
  self->convTmp2 = vl_malloc(sizeof(float) * self->imWidth * self->imHeight) ;

  self->numBinAlloc = 0 ;
  self->numFrameAlloc = 0 ;
  self->numGradAlloc = 0 ;
  self->numCurvAlloc = 0 ;

  self->descrSize = 0 ;
  self->numFrames = 0 ;
  self->grads = NULL ;
  
  self->xxcurv = NULL;
  self->xycurv = NULL;
  self->yxcurv = NULL;
  self->yycurv = NULL;
  
  self->xpixcurv = NULL;
  self->ypixcurv = NULL;
  self->pixcurvmag = NULL;
  
  self->gradmag = NULL;
  
  self->binnedCurv = NULL ;
  
  self->frames = NULL ;
  self->descrs = NULL ;
  self->curvdescrs = NULL ;
  
  self->curvsettings = NULL;

  _vl_curvdsift_update_buffers(self) ;
  return self ;
}

/** ------------------------------------------------------------------
 ** @brief Create a new DSIFT filter (basic interface)
 ** @param imWidth width of the image.
 ** @param imHeight height of the image.
 ** @param step sampling step.
 ** @param binSize bin size.
 ** @return new filter.
 **
 ** The descriptor geometry matches the standard SIFT descriptor.
 **/

VL_EXPORT VlDsiftFilter *
vl_curvdsift_new_basic (int imWidth, int imHeight, int step, int binSize, char *curvsettings)
{
  VlDsiftFilter* self = vl_curvdsift_new(imWidth, imHeight) ;
  VlDsiftDescriptorGeometry geom = *vl_curvdsift_get_geometry(self) ;
  geom.binSizeX = binSize ;
  geom.binSizeY = binSize ;
  vl_curvdsift_set_geometry(self, &geom) ;
  vl_curvdsift_set_steps(self, step, step) ;
  
  if(self->curvsettings) {vl_free(self->curvsettings); self->curvsettings = NULL;}
  self->curvsettings = (char*)vl_malloc(sizeof(char) * 128) ;
  strncpy(self->curvsettings, curvsettings, 127);
  
  return self ;
}

/** ------------------------------------------------------------------
 ** @brief Delete DSIFT filter
 ** @param self DSIFT filter.
 **/

VL_EXPORT void
vl_curvdsift_delete (VlDsiftFilter * self)
{
  _vl_curvdsift_free_buffers (self) ;
  if (self->convTmp2) vl_free (self->convTmp2) ;
  if (self->convTmp1) vl_free (self->convTmp1) ;
  
  if(self->curvsettings) {vl_free(self->curvsettings); self->curvsettings = NULL;}
  
  vl_free (self) ;
}

/** ------------------------------------------------------------------
 ** @brief Export data as binary
 **/

VL_EXPORT void
vl_curvsift_exportdata (char *name, float *data, long len)
{
  FILE * pFile;
  pFile = fopen (name, "wb");
  fwrite (data, sizeof(float), len, pFile);
  fclose (pFile);
}


/** ------------------------------------------------------------------
 ** @internal @brief Process with Gaussian window
 ** @param self DSIFT filter.
 **/

VL_INLINE void
_vl_curvdsift_with_gaussian_window (VlDsiftFilter * self)
{
  int binx, biny, bint ;
  int framex, framey ;
  float *xker, *yker ;

  int Wx = self->geom.binSizeX - 1 ;
  int Wy = self->geom.binSizeY - 1 ;

  for (biny = 0 ; biny < self->geom.numBinY ; ++biny) {

    yker = _vl_curvdsift_new_kernel (self->geom.binSizeY,
                                 self->geom.numBinY,
                                 biny,
                                 self->windowSize) ;

    for (binx = 0 ; binx < self->geom.numBinX ; ++binx) {

      xker = _vl_curvdsift_new_kernel(self->geom.binSizeX,
                                  self->geom.numBinX,
                                  binx,
                                  self->windowSize) ;

      for (bint = 0 ; bint < self->geom.numBinT ; ++bint) {
        // Gaussian window dense convolution (saved to convTmp2)

        vl_imconvcol_vf (self->convTmp1, self->imHeight,
                         self->grads[bint], self->imWidth, self->imHeight,
                         self->imWidth,
                         yker, -Wy, +Wy, 1,
                         VL_PAD_BY_CONTINUITY|VL_TRANSPOSE) ;

        vl_imconvcol_vf (self->convTmp2, self->imWidth,
                         self->convTmp1, self->imHeight, self->imWidth,
                         self->imHeight,
                         xker, -Wx, +Wx, 1,
                         VL_PAD_BY_CONTINUITY|VL_TRANSPOSE) ;
                         
        {
          float *dst = self->descrs
            + bint
            + binx * self->geom.numBinT
            + biny * (self->geom.numBinX * self->geom.numBinT)  ;

          float *src = self->convTmp2 ;

          int frameSizeX = self->geom.binSizeX * (self->geom.numBinX - 1) + 1 ;
          int frameSizeY = self->geom.binSizeY * (self->geom.numBinY - 1) + 1 ;
          int descrSize = vl_curvdsift_get_descriptor_size (self,0) ;

          for (framey  = self->boundMinY ;
               framey <= self->boundMaxY - frameSizeY + 1 ;
               framey += self->stepY) {
            for (framex  = self->boundMinX ;
                 framex <= self->boundMaxX - frameSizeX + 1 ;
                 framex += self->stepX) {
              *dst = src [(framex + binx * self->geom.binSizeX) * 1 +
                          (framey + biny * self->geom.binSizeY) * self->imWidth]  ;
              dst += descrSize ;
            } /* framex */
          } /* framey */
        }

      } /* for bint */
      vl_free (xker) ;
    } /* for binx */
    vl_free (yker) ;
  } /* for biny */
  
  //========== Curvature stuff (BEGIN)

  int curvDescrSize = self->numBinCurv *
                    self->geom.numBinX *
                    self->geom.numBinY ;
  //binnedCurv
  for (biny = 0 ; biny < self->geom.numBinY ; ++biny) {

    yker = _vl_curvdsift_new_kernel (self->geom.binSizeY,
                                 self->geom.numBinY,
                                 biny,
                                 self->windowSize) ;

    for (binx = 0 ; binx < self->geom.numBinX ; ++binx) {

      xker = _vl_curvdsift_new_kernel(self->geom.binSizeX,
                                  self->geom.numBinX,
                                  binx,
                                  self->windowSize) ;

      for (bint = 0 ; bint < self->numBinCurv ; ++bint) {
        // Gaussian window dense convolution (saved to convTmp2)

        vl_imconvcol_vf (self->convTmp1, self->imHeight,
                         self->binnedCurv[bint], self->imWidth, self->imHeight,
                         self->imWidth,
                         yker, -Wy, +Wy, 1,
                         VL_PAD_BY_CONTINUITY|VL_TRANSPOSE) ;

        vl_imconvcol_vf (self->convTmp2, self->imWidth,
                         self->convTmp1, self->imHeight, self->imWidth,
                         self->imHeight,
                         xker, -Wx, +Wx, 1,
                         VL_PAD_BY_CONTINUITY|VL_TRANSPOSE) ;
                         
        {
          float *dst = self->curvdescrs
            + bint
            + binx * self->numBinCurv
            + biny * (self->geom.numBinX * self->numBinCurv)  ;

          float *endDst = self->curvdescrs + self->numCurvAlloc*self->numFrameAlloc;
          
          float *src = self->convTmp2 ;

          /*char nam[32];
          strcpy(nam, "convtmp____.bin");
          nam[8] = 'A' + bint;
          nam[9] = '0' + binx;
          nam[10] = '0' + biny;
          vl_curvsift_exportdata(nam, src, self->imWidth * self->imHeight);*/
          
          int frameSizeX = self->geom.binSizeX * (self->geom.numBinX - 1) + 1 ;
          int frameSizeY = self->geom.binSizeY * (self->geom.numBinY - 1) + 1 ;
          int descrSize = curvDescrSize;

          for (framey  = self->boundMinY ;
               framey <= self->boundMaxY - frameSizeY + 1 ;
               framey += self->stepY) {
            for (framex  = self->boundMinX ;
                 framex <= self->boundMaxX - frameSizeX + 1 ;
                 framex += self->stepX) {
                if(dst >= endDst) {
                  printf("ERR: Exceed: framex %d, framey %d\n", framex, framey);
                  return;
                }
              
              *dst = src [(framex + binx * self->geom.binSizeX) * 1 +
                          (framey + biny * self->geom.binSizeY) * self->imWidth]  ;
              dst += descrSize ;
            } // framex
          } // framey
        }

      } // for bint
      vl_free (xker) ;
    } // for binx
    vl_free (yker) ;
  } // for biny
  
  //========== Curvature stuff (END)
  
}


/** ------------------------------------------------------------------
 ** @brief Normalize a 2D vector field and store normalization factors
 **
 ** @param f1 vector x-component
 ** @param f2 vector y-component
 ** @param norm factors are returned here
 ** @param dim1 size along low-significance dimension (usually y) 
 ** @param dim2 size along high-significance dimension
 **/

void normalize2Dvecfield(float *f1, float *f2, float *norm, int dim1, int dim2)
{
  for(int j = 0; j < dim2; j++) {
    for(int i = 0; i < dim1; i++) {
      int p = j*dim1+i;
      float mag = sqrt(f1[p]*f1[p] + f2[p]*f2[p]);
      norm[p] = mag;
      f1[p] = f1[p] / (mag+eps);
      f2[p] = f2[p] / (mag+eps);
    }
  }
}

/** ------------------------------------------------------------------
 ** @brief Compute keypoints and descriptors
 **
 ** @param self DSIFT filter.
 ** @param im   image data.
 **/

void vl_curvdsift_process (VlDsiftFilter* self, float const* im)
{
  int t, x, y ;
  
    if(self->curvsettings == NULL) {
        printf("Curvsettings string is NULL\n");
        return;
    } else {
        //printf("Curvsettings: %s\n", self->curvsettings);
    }

  /* update buffers */
  _vl_curvdsift_alloc_buffers (self) ;

  /* clear integral images */
  for (t = 0 ; t < self->geom.numBinT ; ++t)
    memset (self->grads[t], 0,
            sizeof(float) * self->imWidth * self->imHeight) ;

  //========== Curvature stuff (BEGIN)
  float hfilter[] = {1,4,6,4,1};
  float vfilter[] = {-1,-2,0,2,1};
  
  //= Soft gradient (5x5 sobel)
  // Note: xpixcurv is used as a buffer for x gradient here! (ypixcurv accordingly)
                        
  vl_pimconv_vf (self->xpixcurv, self->convTmp1, im, self->imWidth, self->imHeight,
                    vfilter, 5, hfilter, 5, 1);
  vl_pimconv_vf (self->ypixcurv, self->convTmp1, im, self->imWidth, self->imHeight,
                    hfilter, 5, vfilter, 5, 1);
  
  //Normalize xgradn, ygradn (temporarily stored in xpixcurv, ypixcurv)
  normalize2Dvecfield(self->xpixcurv, self->ypixcurv, self->gradmag, self->imHeight, self->imWidth); // normalize it
  
  //= xxcurv, xycurv, yxcurv, yycurv:
  vl_pimconv_vf (self->xxcurv, self->convTmp1, self->xpixcurv, self->imWidth, self->imHeight,
                    vfilter, 5, hfilter, 5, 1);
  
  vl_pimconv_vf (self->xycurv, self->convTmp1, self->xpixcurv, self->imWidth, self->imHeight,
                    hfilter, 5, vfilter, 5, 1);
  
  vl_pimconv_vf (self->yxcurv, self->convTmp1, self->ypixcurv, self->imWidth, self->imHeight,
                    vfilter, 5, hfilter, 5, 1);
  
  vl_pimconv_vf (self->yycurv, self->convTmp1, self->ypixcurv, self->imWidth, self->imHeight,
                    hfilter, 5, vfilter, 5, 1);

  float *xgradn = self->xpixcurv;
  float *ygradn = self->ypixcurv;
  
  //========== Curvature stuff (END)

#undef at
#define at(x,y) (im[(y)*self->imWidth+(x)])

  //========== Curvature stuff (BEGIN)
  float magbinthresh[] = {3, 8, 20, 100}; // in corresp. bin, if lower than the threshold (and higher than previous)
  int nummbins = NELEMS(magbinthresh);
  //========== Curvature stuff (END)

  /* Compute gradients, their norm, and their angle */

  int sHasH = strHas(self->curvsettings, 'H'); //Joint histograms?
  
  for (y = 0 ; y < self->imHeight ; ++ y) {
    for (x = 0 ; x < self->imWidth ; ++ x) {
      float gx, gy ;
      float angle, mod, nt, rbint ;
      int bint ;

      /* y derivative */
      if (y == 0) {
        gy = at(x,y+1) - at(x,y) ;
      } else if (y == self->imHeight - 1) {
        gy = at(x,y) - at(x,y-1) ;
      } else {
        gy = 0.5F * (at(x,y+1) - at(x,y-1)) ;
      }

      /* x derivative */
      if (x == 0) {
        gx = at(x+1,y) - at(x,y) ;
      } else if (x == self->imWidth - 1) {
        gx = at(x,y) - at(x-1,y) ;
      } else {
        gx = 0.5F * (at(x+1,y) - at(x-1,y)) ;
      }

      /* angle and modulus */
      angle = vl_fast_atan2_f (gy,gx) ;
      mod = vl_fast_sqrt_f (gx*gx + gy*gy) ;

      /* quantize angle */
      nt = vl_mod_2pi_f (angle) * (self->geom.numBinT / (2*VL_PI)) ;
      bint = (int) vl_floor_f (nt) ;
      rbint = nt - bint ;

      /* write it back */
      self->grads [(bint    ) % self->geom.numBinT][x + y * self->imWidth] = (1 - rbint) * mod ;
      self->grads [(bint + 1) % self->geom.numBinT][x + y * self->imWidth] = (    rbint) * mod ;
      
      //========== Curvature stuff (BEGIN)
      int p = (y)*self->imWidth+(x);
      
      float curv = self->xxcurv[p] * ygradn[p]*ygradn[p]
                + (-self->xycurv[p]-self->yxcurv[p]) * xgradn[p]*ygradn[p]
                + self->yycurv[p] * xgradn[p]*xgradn[p];
                
      self->xpixcurv[p] = curv * xgradn[p];
      self->ypixcurv[p] = curv * ygradn[p];
      self->pixcurvmag[p] = curv;
      
      // Curvature Magnitude Bin:
      int mbin = 0;
      for(;mbin < nummbins; mbin++) {
        if(abs(curv) < magbinthresh[mbin] || mbin == nummbins-1) break;
      }
      
      float dx = self->xpixcurv[p];
      float dy = self->ypixcurv[p];
      float v = self->gradmag[p];
      
      //if(dx == 0 && dy == 0) continue; //invalid pixel
      
      // snap to one of 18 orientations
      double best_dot = 0;
      int best_o = 0;
      for (int o = 0; o < 8; o++) {
        double dot = uu8[o]*dx + vv8[o]*dy;
        if (dot > best_dot) {
          best_dot = dot;
          best_o = o;
        }
      }
      
      if(sHasH) { // Joint (curvH)
        int totalBin = 0;
        if(mbin > 0) {
            totalBin = best_o + 8*(mbin-1) + 1;
        }
        self->binnedCurv[totalBin][x + y * self->imWidth] = v;
        
      } else { // Marginal (curvB)
          
        // Magnitude
        self->binnedCurv[mbin][x + y * self->imWidth] = v;
        // Orientation
        self->binnedCurv[nummbins + best_o][x + y * self->imWidth] = v;
      }
      
      //========== Curvature stuff (END)
    }
  }

  if (self->useFlatWindow) {
	  //Not implemented
	  return;
    //_vl_curvdsift_with_flat_window(self) ;
  } else {
    _vl_curvdsift_with_gaussian_window(self);
  }

  {
    VlDsiftKeypoint* frameIter = self->frames ;
    float * descrIter = self->descrs ;
    int framex, framey, bint ;

    int frameSizeX = self->geom.binSizeX * (self->geom.numBinX - 1) + 1 ;
    int frameSizeY = self->geom.binSizeY * (self->geom.numBinY - 1) + 1 ;
    int descrSize = vl_curvdsift_get_descriptor_size (self,0) ;

    float deltaCenterX = 0.5F * self->geom.binSizeX * (self->geom.numBinX - 1) ;
    float deltaCenterY = 0.5F * self->geom.binSizeY * (self->geom.numBinY - 1) ;

    float normConstant = frameSizeX * frameSizeY ;

    for (framey  = self->boundMinY ;
         framey <= self->boundMaxY - frameSizeY + 1 ;
         framey += self->stepY) {

      for (framex  = self->boundMinX ;
           framex <= self->boundMaxX - frameSizeX + 1 ;
           framex += self->stepX) {

        frameIter->x    = framex + deltaCenterX ;
        frameIter->y    = framey + deltaCenterY ;

        /* mass */
        {
          float mass = 0 ;
          for (bint = 0 ; bint < descrSize ; ++ bint)
            mass += descrIter[bint] ;
          mass /= normConstant ;
          frameIter->norm = mass ;
        }

        /* L2 normalize */
        _vl_curvdsift_normalize_histogram (descrIter, descrIter + descrSize) ;

        /* clamp */
        for(bint = 0 ; bint < descrSize ; ++ bint)
          if (descrIter[bint] > 0.2F) descrIter[bint] = 0.2F ;

        /* L2 normalize */
        _vl_curvdsift_normalize_histogram (descrIter, descrIter + descrSize) ;

        frameIter ++ ;
        descrIter += descrSize ;
      } /* for framex */
    } /* for framey */
  }
  
  //========== Curvature stuff (BEGIN)
  { 
    //printf("Curvature normalization:\n");
    //printf("[%d,%d]\n", self->numCurvAlloc, self->numFrameAlloc);
    
    float * descrIter = self->curvdescrs ;
    float * descrNotExceed = descrIter + (self->numCurvAlloc * self->numFrameAlloc);
    
    
    int framex, framey, bint ;

    int frameSizeX = self->geom.binSizeX * (self->geom.numBinX - 1) + 1 ;
    int frameSizeY = self->geom.binSizeY * (self->geom.numBinY - 1) + 1 ;
    int curvDescrSize = vl_curvdsift_get_descriptor_size (self, 1) ;

    int sHasM = strHas(self->curvsettings, 'M'); //Separate norm for mag and ori
    int sHasN = strHas(self->curvsettings, 'N'); //Joint norm mag and ori
    
    int sHasO = strHas(self->curvsettings, 'O'); //End norm?
    
    int sHasV = strHas(self->curvsettings, 'V'); //Verbose
    
    
    float sNumA = strGetNum(self->curvsettings, 'a'); //clamping after first normalization
    float sNumB = strGetNum(self->curvsettings, 'b'); //end factor
    
    if(sHasV) {
      printf("sHasM,sHasN,sHasO,sNumA,sNumB: [%d,%d,%d,%f,%f]\n", sHasM,sHasN,sHasO,sNumA,sNumB);
    }
    
    for (framey  = self->boundMinY ;
         framey <= self->boundMaxY - frameSizeY + 1 ;
         framey += self->stepY) {

      for (framex  = self->boundMinX ;
           framex <= self->boundMaxX - frameSizeX + 1 ;
           framex += self->stepX) {

        if(descrIter >= descrNotExceed) {
          printf("ERR: Exceed: framex %d, framey %d\n", framex, framey);
          return;
        }
        
        // L2 normalize
        if(sHasM) {
          _vl_curvdsift_normalize_splithistogram (descrIter + 0, descrIter + curvDescrSize, self->numBinCurv, 4) ;
          _vl_curvdsift_normalize_splithistogram (descrIter + 4, descrIter + curvDescrSize, self->numBinCurv, 8) ;
        } else if(sHasN) {
          _vl_curvdsift_normalize_histogram (descrIter + 0, descrIter + curvDescrSize) ;
        }
        
        
        // clamp
        for(bint = 0 ; bint < curvDescrSize ; ++ bint)
          if (descrIter[bint] > sNumA) descrIter[bint] = sNumA;

        if(sHasO) {
          // L2 normalize
          if(sHasM) {
          _vl_curvdsift_normalize_splithistogram (descrIter + 0, descrIter + curvDescrSize, self->numBinCurv, 4) ;
          _vl_curvdsift_normalize_splithistogram (descrIter + 4, descrIter + curvDescrSize, self->numBinCurv, 8) ;
          } else if(sHasN) {
            _vl_curvdsift_normalize_histogram (descrIter + 0, descrIter + curvDescrSize) ;
          }
        }
        
        
        for(bint = 0 ; bint < curvDescrSize ; ++ bint) descrIter[bint] *= sNumB ;

        descrIter += curvDescrSize ;
      } // for framex
    } // for framey
    
  }
  //========== Curvature stuff (END)

}
