/** @file curvsift.h
 ** @brief SIFT (@ref sift)
 ** @author Andrea Vedaldi (mod by Philipp Fischer)
 **/

/*
Copyright (C) 2007-12 Andrea Vedaldi and Brian Fulkerson.
All rights reserved.

This file is part of the VLFeat library and is made available under
the terms of the BSD license (see the COPYING file).
*/

#ifndef VL_CURVSIFT_H
#define VL_CURVSIFT_H

#include <stdio.h>
#include "generic.h"

/** @brief SIFT filter pixel type */
typedef float vl_curvsift_pix ;

/** ------------------------------------------------------------------
 ** @brief SIFT filter keypoint
 **
 ** This structure represent a keypoint as extracted by the SIFT
 ** filter ::VlCurvSiftFilt.
 **/

typedef struct _VlCurvSiftKeypoint
{
  int o ;           /**< o coordinate (octave). */

  int ix ;          /**< Integer unnormalized x coordinate. */
  int iy ;          /**< Integer unnormalized y coordinate. */
  int is ;          /**< Integer s coordinate. */

  float x ;     /**< x coordinate. */
  float y ;     /**< y coordinate. */
  float s ;     /**< s coordinate. */
  float sigma ; /**< scale. */
} VlCurvSiftKeypoint ;

/** ------------------------------------------------------------------
 ** @brief SIFT filter
 **
 ** This filter implements the SIFT detector and descriptor.
 **/

typedef struct _VlCurvSiftFilt
{
  double sigman ;       /**< nominal image smoothing. */
  double sigma0 ;       /**< smoothing of pyramid base. */
  double sigmak ;       /**< k-smoothing */
  double dsigma0 ;      /**< delta-smoothing. */

  int width ;           /**< image width. */
  int height ;          /**< image height. */
  int O ;               /**< number of octaves. */
  int S ;               /**< number of levels per octave. */
  int o_min ;           /**< minimum octave index. */
  int s_min ;           /**< minimum level index. */
  int s_max ;           /**< maximum level index. */
  int o_cur ;           /**< current octave. */

  vl_curvsift_pix *temp ;   /**< temporary pixel buffer. */
  vl_curvsift_pix *octave ; /**< current GSS data. */
  vl_curvsift_pix *dog ;    /**< current DoG data. */
  int octave_width ;    /**< current octave width. */
  int octave_height ;   /**< current octave height. */

  vl_curvsift_pix *gaussFilter ;  /**< current Gaussian filter */
  double gaussFilterSigma ;   /**< current Gaussian filter std */
  vl_size gaussFilterWidth ;  /**< current Gaussian filter width */

  VlCurvSiftKeypoint* keys ;/**< detected keypoints. */
  int nkeys ;           /**< number of detected keypoints. */
  int keys_res ;        /**< size of the keys buffer. */

  double peak_thresh ;  /**< peak threshold. */
  double edge_thresh ;  /**< edge threshold. */
  double norm_thresh ;  /**< norm threshold. */
  double magnif ;       /**< magnification factor. */
  double windowSize ;   /**< size of Gaussian window (in spatial bins) */

  vl_curvsift_pix *grad ;   /**< GSS gradient data. */
  int grad_o ;          /**< GSS gradient data octave. */

  // Those arrays are allocated once to the maximum size and will not be resized acc. to octave:
  vl_curvsift_pix *curvgradmag;   /**< CURV gradient mag. */
  vl_curvsift_pix *curvx;   /**< CURV x. */
  vl_curvsift_pix *curvy;   /**< CURV y. */
  vl_curvsift_pix *curvmag;   /**< CURV mag (=|(x,y)|). */
  vl_curvsift_pix *curvsec;   /**< CURV second order planes (xx,xy,yx,yy). */
  
  int curv_o;

} VlCurvSiftFilt ;

/** @name Create and destroy
 ** @{
 **/
VL_EXPORT
VlCurvSiftFilt*  vl_curvsift_new    (int width, int height,
                             int noctaves, int nlevels,
                             int o_min) ;
VL_EXPORT
void         vl_curvsift_delete (VlCurvSiftFilt *f) ;
/** @} */

/** @name Process data
 ** @{
 **/

VL_EXPORT
int   vl_curvsift_process_first_octave       (VlCurvSiftFilt *f,
                                          vl_curvsift_pix const *im) ;

VL_EXPORT
int   vl_curvsift_process_next_octave        (VlCurvSiftFilt *f) ;

VL_EXPORT
void  vl_curvsift_detect                     (VlCurvSiftFilt *f) ;

VL_EXPORT
int   vl_curvsift_calc_keypoint_orientations (VlCurvSiftFilt *f,
                                          double angles [4],
                                          VlCurvSiftKeypoint const*k);
VL_EXPORT
void  vl_curvsift_calc_keypoint_descriptor   (VlCurvSiftFilt *f,
                                          vl_curvsift_pix *descr,
                                          VlCurvSiftKeypoint const* k,
                                          double angle) ;

VL_EXPORT
void  vl_curvsift_calc_raw_descriptor        (VlCurvSiftFilt const *f,
                                          vl_curvsift_pix const* image,
                                          vl_curvsift_pix *descr,
                                          int widht, int height,
                                          double x, double y,
                                          double s, double angle0) ;

VL_EXPORT
void  vl_curvsift_keypoint_init              (VlCurvSiftFilt const *f,
                                          VlCurvSiftKeypoint *k,
                                          double x,
                                          double y,
                                          double sigma) ;
/** @} */

/** @name Retrieve data and parameters
 ** @{
 **/
VL_INLINE int    vl_curvsift_get_octave_index   (VlCurvSiftFilt const *f) ;
VL_INLINE int    vl_curvsift_get_noctaves       (VlCurvSiftFilt const *f) ;
VL_INLINE int    vl_curvsift_get_octave_first   (VlCurvSiftFilt const *f) ;
VL_INLINE int    vl_curvsift_get_octave_width   (VlCurvSiftFilt const *f) ;
VL_INLINE int    vl_curvsift_get_octave_height  (VlCurvSiftFilt const *f) ;
VL_INLINE int    vl_curvsift_get_nlevels        (VlCurvSiftFilt const *f) ;
VL_INLINE int    vl_curvsift_get_nkeypoints     (VlCurvSiftFilt const *f) ;
VL_INLINE double vl_curvsift_get_peak_thresh    (VlCurvSiftFilt const *f) ;
VL_INLINE double vl_curvsift_get_edge_thresh    (VlCurvSiftFilt const *f) ;
VL_INLINE double vl_curvsift_get_norm_thresh    (VlCurvSiftFilt const *f) ;
VL_INLINE double vl_curvsift_get_magnif         (VlCurvSiftFilt const *f) ;
VL_INLINE double vl_curvsift_get_window_size    (VlCurvSiftFilt const *f) ;

VL_INLINE vl_curvsift_pix *vl_curvsift_get_octave  (VlCurvSiftFilt const *f, int s) ;
VL_INLINE VlCurvSiftKeypoint const *vl_curvsift_get_keypoints (VlCurvSiftFilt const *f) ;
/** @} */

/** @name Set parameters
 ** @{
 **/
VL_INLINE void vl_curvsift_set_peak_thresh (VlCurvSiftFilt *f, double t) ;
VL_INLINE void vl_curvsift_set_edge_thresh (VlCurvSiftFilt *f, double t) ;
VL_INLINE void vl_curvsift_set_norm_thresh (VlCurvSiftFilt *f, double t) ;
VL_INLINE void vl_curvsift_set_magnif      (VlCurvSiftFilt *f, double m) ;
VL_INLINE void vl_curvsift_set_window_size (VlCurvSiftFilt *f, double m) ;
/** @} */

/* -------------------------------------------------------------------
 *                                     Inline functions implementation
 * ---------------------------------------------------------------- */

/** ------------------------------------------------------------------
 ** @brief Get current octave index.
 ** @param f SIFT filter.
 ** @return index of the current octave.
 **/

VL_INLINE int
vl_curvsift_get_octave_index (VlCurvSiftFilt const *f)
{
  return f-> o_cur ;
}

/** ------------------------------------------------------------------
 ** @brief Get number of octaves.
 ** @param f SIFT filter.
 ** @return number of octaves.
 **/

VL_INLINE int
vl_curvsift_get_noctaves (VlCurvSiftFilt const *f)
{
  return f-> O ;
}

/**-------------------------------------------------------------------
 ** @brief Get first octave.
 ** @param f SIFT filter.
 ** @return index of the first octave.
 **/

VL_INLINE int
vl_curvsift_get_octave_first (VlCurvSiftFilt const *f)
{
  return f-> o_min ;
}

/** ------------------------------------------------------------------
 ** @brief Get current octave width
 ** @param f SIFT filter.
 ** @return current octave width.
 **/

VL_INLINE int
vl_curvsift_get_octave_width (VlCurvSiftFilt const *f)
{
  return f-> octave_width ;
}

/** ------------------------------------------------------------------
 ** @brief Get current octave height
 ** @param f SIFT filter.
 ** @return current octave height.
 **/

VL_INLINE int
vl_curvsift_get_octave_height (VlCurvSiftFilt const *f)
{
  return f-> octave_height ;
}

/** ------------------------------------------------------------------
 ** @brief Get current octave data
 ** @param f SIFT filter.
 ** @param s level index.
 **
 ** The level index @a s ranges in the interval <tt>s_min = -1</tt>
 ** and <tt> s_max = S + 2</tt>, where @c S is the number of levels
 ** per octave.
 **
 ** @return pointer to the octave data for level @a s.
 **/

VL_INLINE vl_curvsift_pix *
vl_curvsift_get_octave (VlCurvSiftFilt const *f, int s)
{
  int w = vl_curvsift_get_octave_width  (f) ;
  int h = vl_curvsift_get_octave_height (f) ;
  return f->octave + w * h * (s - f->s_min) ;
}

/** ------------------------------------------------------------------
 ** @brief Get number of levels per octave
 ** @param f SIFT filter.
 ** @return number of leves per octave.
 **/

VL_INLINE int
vl_curvsift_get_nlevels (VlCurvSiftFilt const *f)
{
  return f-> S ;
}

/** ------------------------------------------------------------------
 ** @brief Get number of keypoints.
 ** @param f SIFT filter.
 ** @return number of keypoints.
 **/

VL_INLINE int
vl_curvsift_get_nkeypoints (VlCurvSiftFilt const *f)
{
  return f-> nkeys ;
}

/** ------------------------------------------------------------------
 ** @brief Get keypoints.
 ** @param f SIFT filter.
 ** @return pointer to the keypoints list.
 **/

VL_INLINE VlCurvSiftKeypoint const *
vl_curvsift_get_keypoints (VlCurvSiftFilt const *f)
{
  return f-> keys ;
}

/** ------------------------------------------------------------------
 ** @brief Get peaks treashold
 ** @param f SIFT filter.
 ** @return threshold ;
 **/

VL_INLINE double
vl_curvsift_get_peak_thresh (VlCurvSiftFilt const *f)
{
  return f -> peak_thresh ;
}

/** ------------------------------------------------------------------
 ** @brief Get edges threshold
 ** @param f SIFT filter.
 ** @return threshold.
 **/

VL_INLINE double
vl_curvsift_get_edge_thresh (VlCurvSiftFilt const *f)
{
  return f -> edge_thresh ;
}

/** ------------------------------------------------------------------
 ** @brief Get norm threshold
 ** @param f SIFT filter.
 ** @return threshold.
 **/

VL_INLINE double
vl_curvsift_get_norm_thresh (VlCurvSiftFilt const *f)
{
  return f -> norm_thresh ;
}

/** ------------------------------------------------------------------
 ** @brief Get the magnification factor
 ** @param f SIFT filter.
 ** @return magnification factor.
 **/

VL_INLINE double
vl_curvsift_get_magnif (VlCurvSiftFilt const *f)
{
  return f -> magnif ;
}

/** ------------------------------------------------------------------
 ** @brief Get the Gaussian window size.
 ** @param f SIFT filter.
 ** @return standard deviation of the Gaussian window (in spatial bin units).
 **/

VL_INLINE double
vl_curvsift_get_window_size (VlCurvSiftFilt const *f)
{
  return f -> windowSize ;
}



/** ------------------------------------------------------------------
 ** @brief Set peaks threshold
 ** @param f SIFT filter.
 ** @param t threshold.
 **/

VL_INLINE void
vl_curvsift_set_peak_thresh (VlCurvSiftFilt *f, double t)
{
  f -> peak_thresh = t ;
}

/** ------------------------------------------------------------------
 ** @brief Set edges threshold
 ** @param f SIFT filter.
 ** @param t threshold.
 **/

VL_INLINE void
vl_curvsift_set_edge_thresh (VlCurvSiftFilt *f, double t)
{
  f -> edge_thresh = t ;
}

/** ------------------------------------------------------------------
 ** @brief Set norm threshold
 ** @param f SIFT filter.
 ** @param t threshold.
 **/

VL_INLINE void
vl_curvsift_set_norm_thresh (VlCurvSiftFilt *f, double t)
{
  f -> norm_thresh = t ;
}

/** ------------------------------------------------------------------
 ** @brief Set the magnification factor
 ** @param f SIFT filter.
 ** @param m magnification factor.
 **/

VL_INLINE void
vl_curvsift_set_magnif (VlCurvSiftFilt *f, double m)
{
  f -> magnif = m ;
}

/** ------------------------------------------------------------------
 ** @brief Set the Gaussian window size
 ** @param f SIFT filter.
 ** @param x Gaussian window size (in units of spatial bin).
 **
 ** This is the parameter @f$ \hat \sigma_{\text{win}} @f$ of
 ** the standard SIFT descriptor @ref sift-tech-descriptor-std.
 **/

VL_INLINE void
vl_curvsift_set_window_size (VlCurvSiftFilt *f, double x)
{
  f -> windowSize = x ;
}

/* VL_SIFT_H */
#endif
